// BitStreamReader.cs created with MonoDevelop
// User: morganevans at 11:06 12/02/2009
//
// To change standard headers go to Edit->Preferences->Coding->Standard Headers
//

using System;
using System.IO;

namespace UK_CO_MEVANSPN_libBitStream
{
	/// <summary>
	/// Objects derived from the BitStreamReader class are used to retrieve binary data from a
	/// pre-existing source - currently either a byte array stored in memory or from a file 
	/// stored on read-accessible storage media.
	/// </summary>
	public class BitStreamReader
	{
		/// <summary>
		/// Constructor creates a BitStreamReader object that retrieves data from a byte array in memory.
		/// </summary>
		/// <param name="_data">
		/// Reference to the data block to use.
		/// A <see cref="System.Byte"/>
		/// </param>
		public BitStreamReader(byte[] _data)
		{
			data = _data;
			bytes_to_read = data.Length;
			byte_ptr = 0;
			bit_ptr = 8;
			file_based = false;
			io_failure = false;
		}
		
		/// <summary>
		/// Constructor creates a BitStreamReader bbject that retrieves data from a file.  If the file
		/// cannot be read, the method will throw a System.IO.FileNotFound exception.
		/// </summary>
		/// <param name="_filename">
		/// A string referencing the file holding the BitStreamReader data.
		/// A <see cref="System.String"/>
		/// </param>
		public BitStreamReader(String _filename)
		{
			file_based = true;
			byte_ptr = 0;
			bytes_to_read = 0;
			bit_ptr = 8;
			io_failure = false;
			try {
				in_file = new FileStream(_filename, FileMode.Open);
				readNextBlock();
			} catch (FileNotFoundException fnfe) {
				io_failure = true;
				throw fnfe;
			}
		}
		
		/// <summary>
		/// Returns a byte value defined by the data contained in the number of bits given by the calling method.
		/// </summary>
		/// <param name="_req_bits">
		/// The number of bits to read from the BitStreamReader.  Must be in the range 1 to 8.
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <returns>
		/// The byte value read.
		/// A <see cref="System.Byte"/>
		/// </returns>
		public byte readByte(int _req_bits) {
			if (byte_ptr == bytes_to_read) readNextBlock();
			if (_req_bits < 1 || _req_bits > 8) throw new BitRequirementException(_req_bits);
			byte value = 0;
			if (bit_ptr >= _req_bits) {
				int bit_shift = bit_ptr - _req_bits;
				byte bit_mask = generateBitMask(_req_bits, bit_shift);
				value = (byte) ((data[byte_ptr] & bit_mask) >> bit_shift);
				bit_ptr -= _req_bits;
				checkBounds();
			} else {
				int high_bit_shift = _req_bits - bit_ptr;
				byte high_bits_value = (byte) ((data[byte_ptr] & BitStreamWriter.max_bit_values[bit_ptr - 1]) << high_bit_shift);
				bit_ptr = 8;
				byte_ptr++;
				if (byte_ptr == bytes_to_read) readNextBlock();
				int low_bit_shift = bit_ptr - high_bit_shift;
				byte bit_mask = generateBitMask(high_bit_shift, low_bit_shift);
				byte low_bits_value = (byte) ((data[byte_ptr] & bit_mask) >> low_bit_shift);
				bit_ptr -= high_bit_shift;
				value = (byte) (low_bits_value + high_bits_value);
			}
			return value;
		}
		
		/// <summary>
		/// Returns a char value.  Will throw an EndOfDataException if there is insufficient data left in the
		/// BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// Char value returned from the BitStreamReader data.
		/// A <see cref="System.Char"/>
		/// </returns>
		public char readChar() {
			byte[] char_data = new byte[sizeof(char)];
			for (int i = 0; i < char_data.Length; i++) {
				char_data[i] = readByte(8);
			}
			BinaryReader br = new BinaryReader(new MemoryStream(char_data));
			char val = br.ReadChar();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns a string.  The string is stored in a zero-terminated format.  The method will throw an 
		/// EndOfDataException if there is insufficient data left in the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The string value read.
		/// A <see cref="System.String"/>
		/// </returns>
		public string readString() {
			StringWriter sr = new StringWriter();
			char c = (char) 0;
			do {
				c = readChar();
				if (c != 0) sr.Write(c);
			} while (c != 0);
			
			return sr.ToString();
		}
		
		/// <summary>
		/// Returns an unsigned short value.  The method will throw an EndOfDataException if there is insufficient data 
		/// left in the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The unsigned short value read.
		/// A <see cref="System.UInt16"/>
		/// </returns>
		public ushort readUShort() {
			BinaryReader br = readTypeData(sizeof(ushort));
			ushort val = br.ReadUInt16();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns a short value.  The method will throw an EndOfDataException if there is insufficient data left 
		/// in the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The short value read.
		/// A <see cref="System.UInt16"/>
		/// </returns>
		public short readShort() {
			BinaryReader br = readTypeData(sizeof(short));
			short val = br.ReadInt16();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns an integer value.  The method will throw an EndOfDataException if there is insufficient data left 
		/// in the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The int value read.
		/// A <see cref="System.Int32"/>
		/// </returns>
		public int readInt() {
			BinaryReader br = readTypeData(sizeof(int));
			int val = br.ReadInt32();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns an unsigned integer value.  The method will throw an EndOfDataException if there is insufficient 
		/// data left the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The unsigned integer value read.
		/// A <see cref="System.UInt32"/>
		/// </returns>
		public uint readUInt() {
			BinaryReader br = readTypeData(sizeof(uint));
			uint val = br.ReadUInt32();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns an long value.  The method will throw an EndOfDataException if there is insufficient 
		/// data left the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The long value read.
		/// A <see cref="System.Int64"/>
		/// </returns>
		public long readLong() {
			BinaryReader br = readTypeData(sizeof(long));
			long val = br.ReadInt64();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns an unsigned long value.  The method will throw an EndOfDataException if there is insufficient 
		/// data left the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The unsigned long value read.
		/// A <see cref="System.UInt64"/>
		/// </returns>
		public ulong readULong() {
			BinaryReader br = readTypeData(sizeof(ulong));
			ulong val = br.ReadUInt64();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Returns a float value.  The method will throw an EndOfDataException if there is insufficient 
		/// data left the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The float value read.
		/// A <see cref="System.Single"/>
		/// </returns>
		public float readFloat() {
			BinaryReader br = readTypeData(sizeof(float));
			float val = br.ReadSingle();
			br.Close();
			return val;
		}			

		/// <summary>
		/// Returns a double value.  The method will throw an EndOfDataException if there is insufficient 
		/// data left the BitStreamReader object's data store.
		/// </summary>
		/// <returns>
		/// The double value read.
		/// A <see cref="System.Double"/>
		/// </returns>
		public double readDouble() {
			BinaryReader br = readTypeData(sizeof(double));
			double val = br.ReadDouble();
			br.Close();
			return val;
		}
		
		/// <summary>
		/// Closes the BitStreamReader data storage and hence - access.
		/// </summary>
		public void close() {
			if (file_based) in_file.Close();
			file_based = false;
		}
		
		/// <summary>
		/// Destructor destroys any memory-resident data in the BitStreamData storage and
		/// closes access before deleting the object.
		/// </summary>
		~BitStreamReader() {
			for (int i = 0; i < bytes_to_read; i++) {
				data[i] = 0;
			}
			
			data = null;
			byte_ptr = bit_ptr = -1;
			close();
		}
		
		/// <summary>
		/// Generates a bit mask for internal use.
		/// </summary>
		/// <param name="_req_bits">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <param name="_left_shift">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <returns>
		/// A <see cref="System.Byte"/>
		/// </returns>
		private byte generateBitMask(int _req_bits, int _left_shift) {
			int max_shift = (8 - _req_bits);
			_left_shift = (_left_shift < 0) ? -_left_shift : _left_shift;
			_left_shift = (_left_shift > max_shift) ? max_shift : _left_shift;
			return (byte) (BitStreamWriter.max_bit_values[_req_bits - 1] << _left_shift);
		}
		
		/// <summary>
		/// Internal use only:  Controls read pointer for internal storage.
		/// </summary>
		private void checkBounds() {
			if (bit_ptr <= 0) {
				bit_ptr = 8 + bit_ptr;
				byte_ptr++;
			}
		}
		
		/// <summary>
		/// Internal use:  Reads the appropriate number of bits for a particular type from storage.
		/// </summary>
		/// <param name="_bytes">
		/// A <see cref="System.Int32"/>
		/// </param>
		/// <returns>
		/// A <see cref="BinaryReader"/>
		/// </returns>
		private BinaryReader readTypeData(int _bytes) {
			byte[] type_data = new byte[_bytes];
			for (int i = 0; i < type_data.Length; i++)
				type_data[i] = readByte(8);
			return new BinaryReader(new MemoryStream(type_data));
		}
		
		/// <summary>
		/// Internal use:  For file-based storage will read the next block of storage data 
		/// when the current block is exhausted.
		/// </summary>
		private void readNextBlock() {
			if (io_failure || !file_based) throw new EndOfDataException();
			data = new byte[4096];
			bytes_to_read = in_file.Read(data, 0, 4096);
			if (bytes_to_read == 0) {
				in_file.Close();
				io_failure = true;
				throw new EndOfDataException();
			}
		}
		
		private byte[] data;
		private int byte_ptr, bit_ptr, bytes_to_read;
		private bool file_based, io_failure;
		private FileStream in_file;
	}
}
