// BitStreamWriter.cs created with MonoDevelop
// User: morganevans at 18:11 11/02/2009
//
// To change standard headers go to Edit->Preferences->Coding->Standard Headers
//

using System;
using System.IO;
using System.Collections;

namespace UK_CO_MEVANSPN_libBitStream {
	/// <summary>
	/// Objects derived from this class are used to write data of arbitrary length to a
	/// stream whose data can later be retrieved as a memory-resident byte array or can 
	/// be written directly to file on writable storage media.
	/// </summary>
	public class BitStreamWriter {
		/// <summary>
		/// This static constructor is used internally to define some data structure relevent
		/// to the class method's internal operation.
		/// </summary>
		static BitStreamWriter() {
			max_bit_values = new byte[8];
			for (int i = 0; i < 8; i++) {
				int bit_value = 1;
				max_bit_values[i] = 0;
				for (int j = 0; j <= i; j++) {
					max_bit_values[i] += (byte) bit_value;
					bit_value = bit_value * 2;
				}
			}
		}
		
		/// <summary>
		/// Creates a BitStreamWriter using the default sub-storage block size of 4096 bytes.
		/// This constructor is meant for general purpose usage but may be less efficient than 
		/// the secondary constructor which allows the sub-storage block size to be user-defined, allowing
		/// for larger blocks more suited for efficient storage with larger data storage requirements.
		/// </summary>
		public BitStreamWriter() {
			data_blocks = new ArrayList();
			new_block_size = 4096;
			createDataBlock();
			for (int i = 0; i < data_ptr.Length; i++) {
				data_ptr[i] = 0;
			}
			byte_ptr = bit_ptr = 0;
		}
		
		/// <summary>
		/// Createa a BitStreamWriter object with a user-defined sub-block storage size (in bytes)  Larger sub-blocks 
		/// will make the storage routines more efficient whilst small sub-blocks are suitable for compact data
		/// management where memory is limited.  The default constructor uses a sub-block size of 4096 bytes.
		/// </summary>
		/// <param name="_block_size">
		/// Size of the sub-block storage in bytes.
		/// A <see cref="System.Int32"/>
		/// </param>
		public BitStreamWriter(int _block_size) {
			new_block_size = (_block_size < 1) ? 4096 : _block_size;
			data_blocks = new ArrayList();
			createDataBlock();
			for (int i = 0; i < data_ptr.Length; i++) {
				data_ptr[i] = 0;
			}
			byte_ptr = bit_ptr = 0;
		}
		
		/// <summary>
		/// Writes a byte value to the BitStreamWriter storage, fitting it into the required number of bits.
		/// If the value is too large to fit into the required number of bits or the required bit-length is 
		/// less than 1 or greater that 8 then a BitRequirementException is thrown.
		/// </summary>
		/// <param name="_value">
		/// The value to write to storage.
		/// A <see cref="System.Byte"/>
		/// </param>
		/// The number of bits to use to store the value.
		/// <param name="_req_bits">
		/// A <see cref="System.Int32"/>
		/// </param>
		public void write(byte _value, int _req_bits) {
			int value_bit_count = getBitCount(_value);
			if (value_bit_count < _req_bits) {
				bit_ptr += _req_bits - value_bit_count;
				if (bit_ptr >= 8) {
					bit_ptr = bit_ptr - 8;
					byte_ptr++;
					checkStorage();
				}
				_req_bits = value_bit_count;
			} else {
				_req_bits = (_req_bits < 0) ? -_req_bits : _req_bits;
			}
			if (_req_bits == 0 || _req_bits > 8 || _value > max_bit_values[_req_bits - 1]) throw new BitRequirementException(_req_bits);
			int bits_remaining = 8 - bit_ptr;
			if (bits_remaining > _req_bits) {
				int shift_left = (int) (bits_remaining - _req_bits);
				int shifted_value = _value << shift_left;
				data_ptr[byte_ptr] = (byte) (data_ptr[byte_ptr] | shifted_value);
				bit_ptr += _req_bits;
				if (bit_ptr == 8) {
					byte_ptr++;
					bit_ptr = 0;
					checkStorage();
				}
			} else {
				int bit_shift = (int) _req_bits - bits_remaining;
				byte high_bits_byte = (byte) (_value >> bit_shift);
				byte low_bits_byte = (byte) (_value << (8 - bit_shift));
				data_ptr[byte_ptr] = (byte) (data_ptr[byte_ptr] | high_bits_byte);
				byte_ptr++;
				checkStorage();
				data_ptr[byte_ptr] = low_bits_byte;
				bit_ptr = bit_shift;
			}
		}
		
		/// <summary>
		/// Writes an array of bytes to storage using the default bit length of 8-bits per byte.
		/// </summary>
		/// <param name="_bytes">
		/// The byte array to write to storage.
		/// A <see cref="System.Byte"/>
		/// </param>
		public void writeBytes(byte[] _bytes) {
			for (int i = 0; i < _bytes.Length; i++) write(_bytes[i], 8);
		}
		
		/// <summary>
		/// Writes a char value to storage.
		/// </summary>
		/// <param name="_value">
		/// Char value to write.
		/// A <see cref="System.Char"/>
		/// </param>
		public void write(char _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes an chort value to storage.
		/// </summary>
		/// <param name="_value">
		/// Short value to write.
		/// A <see cref="System.Int16"/>
		/// </param>
		public void write(short _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes an unsigned short value to storage.
		/// </summary>
		/// <param name="_value">
		/// Value to write.
		/// A <see cref="System.UInt16"/>
		/// </param>
		public void write(ushort _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes an integer value to storage.
		/// </summary>
		/// <param name="_value">
		/// Value to write.
		/// A <see cref="System.Int32"/>
		/// </param>
		public void write(int _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes an unsigned integer value to storage.
		/// </summary>
		/// <param name="_value">
		/// Value to write.
		/// A <see cref="System.UInt32"/>
		/// </param>
		public void write(uint _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes a float value to storage.
		/// </summary>
		/// <param name="_value">
		/// Write to write.
		/// A <see cref="System.Single"/>
		/// </param>
		public void write(float _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes a double precision float value to storage.
		/// </summary>
		/// <param name="_value">
		/// Value to write.
		/// A <see cref="System.Double"/>
		/// </param>
		public void write(double _value) {
			byte[] b = BitConverter.GetBytes(_value);
			writeBytes(b);
		}
		
		/// <summary>
		/// Writes a string to storage.  Note that the string is stored as an array of char values, not bytes.
		/// </summary>
		/// <param name="_value">
		/// String to write.
		/// A <see cref="System.String"/>
		/// </param>
		public void write(string _value) {
			char[] str_chars = _value.ToCharArray();
			for (int i = 0; i < str_chars.Length; i++) write(str_chars[i]);
			write((char) 0);
		}
		
		/// <summary>
		/// Returns the stored values as a byte array.
		/// </summary>
		/// <returns>
		/// An array of bytes representing the stored data.
		/// A <see cref="System.Byte"/>
		/// </returns>
		public byte[] getBytes() {
			// Count the size of all data blocks;
			int data_len = (data_blocks.Count * new_block_size) + ((bit_ptr > 0) ? 1 : 0);
			byte[] return_data = new byte[data_len];
			int wp = 0;
			for (int i = 0; i < data_blocks.Count; i++) {
				int block_len = (i != data_blocks.Count - 1) ? new_block_size : byte_ptr + ((bit_ptr > 0) ? 1 : 0);
				byte[] block = (byte[]) data_blocks[i];
				for (int j = 0; j < block_len; j++) {
					return_data[wp++] = block[j];
				}
			}
			
			return return_data;
		}
		
		/// <summary>
		/// Saves the stored data directly to file.  If the filename given points to an inaccessible storage location
		/// an exception will be thrown.
		/// </summary>
		/// <param name="_filename">
		/// Filename (including optional path) used to store this data.
		/// A <see cref="System.String"/>
		/// </param>
		public void saveToFile(string _filename) {
			FileStream fs = new FileStream(_filename, FileMode.Create);
			for (int i = 0; i < data_blocks.Count; i++) {
				byte[] file_block = (byte[]) data_blocks[i];
				int block_len = (i != data_blocks.Count - 1) ? new_block_size : byte_ptr + ((bit_ptr > 0) ? 1 : 0);
				fs.Write(file_block, 0, block_len);
			}
			fs.Close();
		}
		
		/// <summary>
		/// Utility method:  Returns the minimum number of bits required to store a byte value.
		/// </summary>
		/// <param name="_value">
		/// Value to store.
		/// A <see cref="System.Byte"/>
		/// </param>
		/// <returns>
		/// Minimum number of bits required to store the value.
		/// A <see cref="System.Int32"/>
		/// </returns>
		public static int getBitCount(byte _value) {
			int i;
			for (i = 0; i < 8 && _value > max_bit_values[i]; i++);
			return i + 1;
		}
		
		/// <summary>
		/// Internal method.  Controls sub-block generation.
		/// </summary>
		private void checkStorage() {
			int data_len = data_ptr.Length;
			if (byte_ptr == data_len) createDataBlock();
		}
		
		/// <summary>
		/// Internal method.  Createa sub-block storage.
		/// </summary>
		private void createDataBlock() {
			data_ptr = new byte[new_block_size];
			data_blocks.Add(data_ptr);
			byte_ptr = 0;
			
		}
		
		/// <summary>
		/// Destructor will overwrite all data in storage with null values and clean up before removing the object.
		/// </summary>
		~BitStreamWriter() {
			for (int i = 0; i < data_blocks.Count; i++) {
				byte[] block = (byte[]) data_blocks[i];
				for (int j = 0; j < new_block_size; j++) {
					block[j] = 0;
				}
			}
			
			data_blocks.Clear();
			data_ptr = null;
			byte_ptr = bit_ptr = -1;
		}
		
		private byte[] data_ptr;
		private int byte_ptr, bit_ptr;
		private ArrayList data_blocks;
		private int new_block_size;
		
		/// <summary>
		/// Utility membe: Defines an array of values representing the maximum value storable in a particular bit length.
		/// </summary>
		public static readonly byte[] max_bit_values;
	}
}